package com.sharecare.core;

import lombok.RequiredArgsConstructor;
import lombok.Value;
import org.apache.tomcat.util.buf.HexUtils;
import org.springframework.stereotype.Service;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.Charset;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.util.stream.Collectors;

/**
 * Service class to deal with the creation of users and
 * insert them into MariaDB
 *
 * Uses Hash&Salt for a secure registration and login.
 */
@Service
@RequiredArgsConstructor
public class UserService {
    private final UserRepository userRepository;
    private final static SecureRandom SECURE_RANDOM;
    static {
        try {
            SECURE_RANDOM = SecureRandom.getInstance("SHA1PRNG");
        } catch (NoSuchAlgorithmException e) {
            throw new RuntimeException(e);
        }
    }

    public User createUser(String name, String email, CharSequence password) {
        User u = new User();
        u.setName(name);
        u.setEmail(email);
        Password p = createPassword(toChar(password));
        u.setPasswordHash(p.getHash());
        u.setPasswordSalt(p.getSalt());
        u = userRepository.save(u);
        return u;
    }

    public boolean isAuthorized(String email, CharSequence password) {
        User u = userRepository.findByEmail(email);
        if (u == null) {
            return false;
        }
        String salt = u.getPasswordSalt();
        String thisHash = String.copyValueOf(hash(toChar(password), toChar(salt)));
        return u.getPasswordHash().equals(thisHash);
    }

    public char[] hash(char[] c1, char[] c2) {
        return hash(concat(c1, c2));
    }

    public Password createPassword(char[] password) {
        char[] salt = createSalt();
        char[] hash = hash(password, salt);
        Password p = new Password(String.copyValueOf(hash), String.copyValueOf(salt));
        return p;
    }

    private char[] toChar(CharSequence s) {
        char[] c = new char[s.length()];
        for (int i = 0; i < s.length(); i++) {
            c[i] = s.charAt(i);
        }
        return c;
    }

    private char[] hash(char[] concat) {
        byte[] bytes = toBytes(concat);
        MessageDigest messageDigest = null;
        try {
            messageDigest = MessageDigest.getInstance("SHA-256");
        } catch (NoSuchAlgorithmException e) {
            throw new RuntimeException(e);
        }
        bytes = messageDigest.digest(bytes);
        return toChars(bytes);
    }

    private char[] toChars(byte[] bytes) {
        return HexUtils.toHexString(bytes).toCharArray();
    }

    private byte[] toBytes(char[] concat) {
        ByteBuffer bb = Charset.forName("UTF-8").encode(CharBuffer.wrap(concat));
        byte[] b = new byte[bb.limit()];
        bb.get(b);
        return b;

    }


    private char[] createSalt() {
        String s = SECURE_RANDOM.ints(32, 0, LANDN.length())
                .mapToObj(LANDN::charAt)
                .map(String::valueOf)
                .collect(Collectors.joining(""));
        return s.toCharArray();
    }
    private static final String LANDN="ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890";

    private char[] concat(char[] c1, char[] c2) {
        char[] a  = new char[c1.length + c2.length];
        for (int i = 0; i < c1.length; i++)  {
            a[i] = c1[i];
        }
        for (int i = 0; i < c2.length; i++) {
            a[i + c1.length] = c2[i];
        }
        return a;
    }


    @Value
    private static class Password {
        private final String hash;
        private final String salt;
    }

}
