package com.sharecare.core.like;

import com.sharecare.core.Question;
import com.sharecare.core.User;
import lombok.RequiredArgsConstructor;
import org.springframework.stereotype.Service;
import redis.clients.jedis.Jedis;

import java.util.Optional;

/**
 * Like system. Stores the state of "likeness" of a question in the specific user,
 * and keeps track of the likes of a question in Redis
 */
@Service
@RequiredArgsConstructor
public class LikeService {

    private final Jedis jedis;

    public void addLike(User user, Question question) {
        setVote(user, question, Vote.LIKE);
    }

    public void addDislike(User user, Question question) {
        setVote(user, question, Vote.DISLIKE);
    }

    public void undoLikeOrDislike(User user, Question question) {
        unsetVote(user, question);
    }

    public Optional<Vote> getVote(User user, Question question) {
        return Vote.fromIntegerString(jedis.get(getKey(user, question)));
    }


    public long getLikes(Question question) {
        return Long.valueOf(jedis.get(getLikeCountKey(question)));
    }

    public long getDislikes(Question q) {
        return Long.valueOf(jedis.get(getDislikeCountKey(q)));
    }


    private void setVote(User user, Question question, Vote vote) {
        jedis.set(getKey(user, question), vote.getStringValue());
        if (vote.isLike()) {
            jedis.incr(getLikeCountKey(question));
        } else {
            jedis.incr(getDislikeCountKey(question));
        }
    }

    private void unsetVote(User user, Question question) {
        getVote(user, question).ifPresent(vote -> {
            if (vote.isLike()) {
                jedis.decr(getLikeCountKey(question));
            } else {
                jedis.decr(getDislikeCountKey(question));
            }
            jedis.del(getKey(user, question));
        });
    }

    private String getKey(User user, Question question) {
        return "like:" + user.getId() + "-" + question.getId();
    }

    private String getDislikeCountKey(Question question) {
        return "dislikeQuestionCount:" + question.getId();
    }

    private String getLikeCountKey(Question question) {
        return "likeQuestionCount:" + question.getId();
    }

    public void initLikes(Question question) {
        jedis.set(getLikeCountKey(question), "0");
        jedis.set(getDislikeCountKey(question), "0");
    }
}
