package com.sharecare.core.searchable;

import com.sharecare.core.*;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.client.Client;
import org.elasticsearch.index.query.QueryBuilders;
import org.springframework.context.ApplicationContext;
import org.springframework.data.jpa.repository.JpaRepository;
import org.springframework.stereotype.Service;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

@Slf4j
@Service
@RequiredArgsConstructor
public class SearchService {

    private final Client client;
    private final ApplicationContext applicationContext;

    public List<SearchableEntity> searchEntities(String searchText, int from) {
        List<SearchableContent> contents = searchContents(searchText, from);
        return contents.stream()
            .map(this::toEntity)
            .collect(Collectors.toList());
    }

    //todo: Careful! Might happen that an entity was deleted, but still exists in ES! I need to delete it too!
    private SearchableEntity toEntity(SearchableContent searchableContent) {

        /*System.out.println("Bean names: " +
                Arrays.toString(applicationContext.getBeanNamesForType(QuestionRepository.class)));
        System.out.println("Bean names: " +
                Arrays.toString(applicationContext.getBeanNamesForType(AnswerRepository.class)));*/
        Object repository = applicationContext.getBean(searchableContent.getClassName().toLowerCase() + "Repository");
        if (repository == null) {
            log.error("Could not find repository for searchable content class {}. Create a repository with bean name {}", searchableContent.getClassName(), searchableContent.getClassName() + "Repository" );
            throw new IllegalArgumentException();
        }
        if (!(repository instanceof JpaRepository)) {
            throw new IllegalArgumentException();
        } else {
            JpaRepository jpaRepository = (JpaRepository) repository;
            Object entity = jpaRepository.findOne(searchableContent.getEntityId());
            if (!(entity instanceof SearchableEntity)) {
                throw new IllegalArgumentException();
            } else {
                return (SearchableEntity) entity;
            }
        }
    }

    //Analyzer for spanish language has been added to field "all_content" during
    //the creation of the index. Check index creation document
    public List<SearchableContent> searchContents(String searchText, int from) {
        SearchResponse searchResponse = client.prepareSearch("questions")
                .setFrom(from)
                .setQuery(QueryBuilders.matchQuery("all_content", searchText))
                .get();

        return Arrays.stream(searchResponse.getHits().getHits())
                .map(hit -> SearchableContent.builder()
                        .body(hit.getSourceAsMap().get("body").toString())
                        .title(hit.getSourceAsMap().get("title").toString())
                        .className(hit.getSourceAsMap().get("className").toString())
                        .entityId(Long.valueOf(hit.getSourceAsMap().get("entityId").toString()))
                        /*.body(hit.getFields().get("body").getValue())
                        .title(hit.getFields().get("title").getValue())
                        .className(hit.getFields().get("className").getValue())
                        .entityId(hit.getFields().get("entityId").getValue())*/
                .build())
                .collect(Collectors.toList());
    }

    public void makeSearchable(SearchableEntity searchableEntity) {
        SearchableContent.SearchableContentBuilder builder = SearchableContent.builder();

        builder.className(searchableEntity.getClass().getSimpleName());

        builder.entityId(searchableEntity.getId());
        builder.body(searchableEntity.getText());

        if (searchableEntity instanceof WithTitleEntity) {
            builder.title(((WithTitleEntity) searchableEntity).getTitle());
        } else {
            builder.title("");
        }
        save(builder.build());
    }

    //todo: maybe change type according to the type of the doc, or maybe use generic type for all kinds of body texts
    private void save(SearchableContent searchableContent) {
        client.prepareIndex()
                .setIndex("questions")
                .setType("question")
                .setSource(toMap(searchableContent))
                .execute();//Makes the operation asynchronous.
    }

    //Query to delete the corresponding item in ES. Assumes entityId+className+body
    //would make enough of a difference to only find one match
    public void delete(SearchableEntity searchableEntity) {

        SearchResponse searchResponse = client.prepareSearch("questions")
                .setQuery(QueryBuilders.matchQuery("entityId", searchableEntity.getId()))
                .setQuery(QueryBuilders.matchQuery("className", searchableEntity.getClass()))
                .setQuery(QueryBuilders.matchQuery("body", searchableEntity.getText()))
                .get();

        //searchResponse.getHits().getHits()[0].getId()

        client.prepareDelete()
                .setIndex("questions")
                .setType("question")
                .setId(searchResponse.getHits().getHits()[0].getId())
                .get();
    }

    private Map<String, Object> toMap(SearchableContent searchableContent) {
        Map<String, Object> m = new HashMap<>();
        m.put("body", searchableContent.getBody());
        m.put("title", searchableContent.getTitle());
        m.put("className", searchableContent.getClassName());
        m.put("entityId", searchableContent.getEntityId());
        m.put("all_content", searchableContent.getBody() + " " + searchableContent.getTitle());
        return m;
    }

}
